// name: AtomPositions.C
// author: J. Michael Word
// date written: 8/1/97
// purpose: Implementation for AtomPositions

// **************************************************************
// NOTICE: This is free software and the source code is freely
// available. You are free to redistribute or modify under the
// conditions that (1) this notice is not removed or modified
// in any way and (2) any modified versions of the program are
// also available for free.
//               ** Absolutely no Warranty **
// Copyright (C) 1999 J. Michael Word
// **************************************************************

#ifdef OLD_STD_HDRS
#include <stdio.h>
#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>
#include <stdlib.h>
#include <ctype.h>
#else
#include <cstdio>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <cstdlib>
#include <cctype>
using std::ifstream;
using std::cout;
using std::cerr;
using std::endl;
using std::ios_base;
#endif


#include "AtomPositions.h"
#include "Vector.h"
#include "DisjointSets.h"
#include "StdResH.h"
#include "RotMethyl.h"
#include "RotDonor.h"
#include "FlipMemo.h"

int AtomPositions::forceOrientations(const Stringclass& ofilename, List<Stringclass>& orientNotes) {
   int numInNonDefaultOrientations = 0;
   ifstream is(ofilename);
   if (! is) {
      cerr << "ERROR in Fix: can't open orientation file \""
           << ofilename << "\"." << endl;
      return 0;
   }

   const int buflen = 200;
   char buf[buflen+1];

   while (is.getline(buf, buflen)) {
      char orient = ' ';
      float angle = 0.0;
      int oval    = 0;
      char *p = buf;
      for (; *p && (p[0] != ':') && (p[0] != '#'); p++) {
	 if (orient == ' ') {
	    if (toupper(*p) == 'O') { orient = 'O'; }
	    if (toupper(*p) == 'F') { orient = 'F'; }
	    if (toupper(*p) == 'R') { orient = 'R'; }
	 }
	 else if (orient == 'F') {
	    if (*p == '+' || *p == '-' || isdigit(*p)) {
	       oval = atoi(p);
	       while (*p && (p[0] != ':')) { p++; }
	       break;
	    }
	 }
	 else if (orient == 'R') {
	    if (*p == '+' || *p == '-' || isdigit(*p)) {
	       angle = atof(p);
	       while (*p && (p[0] != ':')) { p++; }
	       break;
	    }
	 }
      }
      Stringclass descr;
      if (p[0] == ':') {
	 char *q = ++p;
	 while (*p && (p[0] != ':')) { p++; }
	 p[0] = '\0';
	 descr = Stringclass(q);
      }
      Mover *mx;
      if (descr.empty()) { continue; } // incomplete line or comment
      else {
	 mx = _motionDesc.get(descr);
      }
      if (mx != NULL && mx->valid() /* && mx->isComplete() */ ) {
	 if (orient == 'O' && (mx->canRotate()
	      || (mx->canFlip() && (mx->numOrientations() <= 2)) )) {
	    if (_outputNotice) {
	       cerr << "FIX orientation=ORIG:" <<descr<< ":" << endl;
	    }
	    ::sprintf(buf, ( "USER  MOD Fix    :%s:%s: (o)" ),
	       mx->descr().array(),(const char*)(mx->describeOrientation()));
	    orientNotes.append(Stringclass(buf));
	    mx->makeNonAdjustable();
	 }
	 else if (orient == 'O' && mx->canFlip()) {
	    mx->limitOrientations(FALSE);
	    if (_outputNotice) {
	       cerr << "LIM orientation=ORIG:" <<descr<< ": "
	            << mx->numOrientations() << " states" << endl;
	    }
	    ::sprintf(buf, ( "USER  MOD Limit  :%s:           : (o, %d states)" ),
	       mx->descr().array(),
	       mx->numOrientations());
	    orientNotes.append(Stringclass(buf));
	 }
	 else if (orient == 'F'
	      && mx->canFlip() && (mx->numOrientations() <= 2)) {

	    mx->setOrientation(1, *this);

	    if (_outputNotice) {
	       cerr << "FIX orientation=FLIP:" <<descr<< ":" << endl;
            }
	    ::sprintf(buf, ( "USER  MOD Fix    :%s:%s: (f)" ),
	       mx->descr().array(),(const char*)(mx->describeOrientation()));
	    orientNotes.append(Stringclass(buf));
	    mx->makeNonAdjustable();
	    numInNonDefaultOrientations++;
	 }
	 else if (orient == 'F' && mx->canFlip()) {
	    if ((oval > 0) && (oval <= mx->numOrientations())) {
	       mx->setOrientation(oval-1, *this);

	       if (_outputNotice) {
	          cerr << "FIX orientation=  "<<oval<<" :" <<descr<< ":" << endl;
               }
	       ::sprintf(buf, ( "USER  MOD Fix    :%s:%s: (f %d)" ),
		  mx->descr().array(),
		  (const char*)(mx->describeOrientation()), oval);
	       orientNotes.append(Stringclass(buf));
	       mx->makeNonAdjustable();
	       numInNonDefaultOrientations++;
	    }
	    else if (oval == 0) {
	       mx->limitOrientations(TRUE);

	       if (_outputNotice) {
	          cerr << "LIM orientation=FLIP:" <<descr<< ": "
	               << mx->numOrientations() << " states" << endl;
	       }
	       ::sprintf(buf, ( "USER  MOD Limit  :%s:FLIP       : (f, %d states)" ),
		  mx->descr().array(),
		  mx->numOrientations());
	       orientNotes.append(Stringclass(buf));
	    }
	    else {
	       cerr << "ERROR in Fix:" << descr
		    << ": can't set orientation to "
		    << oval << ", range = 1.."
		    << mx->numOrientations() << endl;
	    }
	 }
	 else if (orient == 'R' && mx->canRotate()) {
	    mx->setHydAngle(clampAngle(angle), *this);
	    if (_outputNotice) {
	       cerr << "FIX rot ang="<<clampAngle(angle)
	            << " deg :" <<descr<< ":" << endl;
	    }
	    ::sprintf(buf, ( "USER  MOD Fix    :%s:%s: (r %.0f)" ),
		  mx->descr().array(),
		  (const char*)(mx->describeOrientation()), angle);
	    orientNotes.append(Stringclass(buf));
	    mx->makeNonAdjustable();
	    numInNonDefaultOrientations++;
	 }
	 else {
	    cerr << "ERROR in Fix:" << descr << ": don't know how to '"
		 << buf << "'" << endl;
	 }
      }
      else { cerr << "ERROR in Fix:" << descr
                  << ": not a valid adjustable group." << endl; }
   }
   return numInNonDefaultOrientations;
}

// ---------------------------------------------------------------
// move atom to new xyz pos.

void AtomPositions::reposition(const Point3d& prev, const PDBrec& rec) {
   LocBlk pb(prev);
   LocBlk nb(rec.loc());
   if (! (pb == nb)) { // rec has moved to a new xyz block
      Seq<PDBrec>a = _xyzBlocks.get(pb);
      if (a.len() > 0) { _xyzBlocks.drop(pb); } // rebuild the old block
      for (; a; a++) {
	 if ((*a).valid() && ! ((*a) == rec)) { _xyzBlocks.put(pb, *a); }
      }
      _xyzBlocks.put(nb, rec); // add to new block
   }
}

// ---------------------------------------------------------------
void AtomPositions::insertRot(const PDBrec& hr,
                const PDBrec& c1, const PDBrec& c2, const PDBrec& c3,
		bool doOHSH, bool doNH3, bool doMethyl) {

   if ((! visableAltConf(hr, _onlyA))
    || (! visableAltConf(c1, _onlyA))
    || (! visableAltConf(c2, _onlyA))) { return; }

   char descrbuf[30];

   ::sprintf(descrbuf, "%c%4d%c%-3.3s%-4.4s%c",
	    hr.chain(), hr.resno(), hr.insCode(),
	    hr.resname(), c1.atomname(), hr.alt());
   Stringclass descr = Stringclass(descrbuf);

   const double dang = dihedral(hr.loc(), c1.loc(), c2.loc(), c3.loc());

// cerr << "AtomPositions::insertRot(dang=" << dang << ") "
//      << hr.chain() << " " <<  hr.resno() << hr.insCode()
//      << " " << hr.resname() << " " << hr.atomname() <<  hr.alt() << endl;

   Mover *m = _motionDesc.get(descr);
   if (m == NULL) {
      if (c1.elem().atno() == 6) {
	 if (doMethyl) {
	    m = (Mover*)new RotMethyl(c1.loc(), c2.loc(), dang, c1);
	    _motionDesc.put(descr, m);
	 }
      }
      else if (c1.elem().atno() == 7) {
	 if (doNH3) {
	    m = (Mover*)new RotMethyl(c1.loc(), c2.loc(), dang, c1);
	    _motionDesc.put(descr, m);
	 }
      }
      else if (c1.hasProp(ACCEPTOR_ATOM)) {
	 if (doOHSH) {
	    m = (Mover*)new RotDonor(c1.loc(), c2.loc(), dang, c1);
	    _motionDesc.put(descr, m);
	 }
      }
      else {
	 cerr<<"*error* insertRot(heavy atom("<<
	    PDBrecNAMEout(c1)<<") not Carbon and not ACCEPTOR)"<<endl;
      }
   }
   if (m != NULL) {
      if (m->type() == Mover::ROTATE_METHYL){
	 if (doMethyl || doNH3) {
	    ((RotMethyl*)m)->insertHatom(hr);
	 }
      }
      else if (m->type() == Mover::ROTATE_DONOR){
	 if (doOHSH) {
	    ((RotDonor*)m)->insertHatom(hr);
	 }
      }
      else {
	 cerr<<"*error* insertRot(hr, "<< m->type() <<")"<<endl;
      }
   }
}

// ---------------------------------------------------------------
// Make the group non-adjustable (fixed)

void AtomPositions::doNotAdjust(const PDBrec& a) {
   char descrbuf[30];

   ::sprintf(descrbuf, "%c%4d%c%-3.3s%-4.4s%c",
		  a.chain(), a.resno(), a.insCode(),
		  a.resname(), "", a.alt());
   const Stringclass descr = Stringclass(descrbuf);

   Mover *m = _motionDesc.get(descr);
   if (m != NULL) {
      m->makeNonAdjustable();
   }
}

// ---------------------------------------------------------------
// Insert the atoms in a resdue block into one or more flip memos
// and return the alt codes for the memos (empty if res not flipped)

Seq<char> AtomPositions::insertFlip(const ResBlk& rblk) {
   Seq<char> resAlts = FlipMemo::altCodes(rblk, _useXplorNames);
   MultiDictIterator<Stringclass, PDBrec> pdbit = rblk.atomIt();
   Stringclass key, descr;
   char descrbuf[30];

   for(Seq<char> alts = resAlts; alts; alts++) {
      pdbit.reset();
      while(pdbit.next(key)) {
	 if (key[0] == '-' || key[0] == '+') { continue; } // skip connectors

	 for (Seq<PDBrec> atsq = rblk.get(key); atsq; atsq++) {
	    if ( visableAltConf(*atsq, _onlyA)
	    &&  ((*atsq).alt() == ' ' || (*atsq).alt() == *alts) ) {

	       ::sprintf(descrbuf, "%c%4d%c%-3.3s%-4.4s%c",
		  (*atsq).chain(), (*atsq).resno(), (*atsq).insCode(),
		  (*atsq).resname(), "", (*alts));
	       descr = Stringclass(descrbuf);

	       Mover *m = _motionDesc.get(descr);
	       if (m == NULL) {
		  m = new FlipMemo((*atsq).resname(), _useXplorNames);
		  _motionDesc.put(descr, m);
	       }
	       else if (m->type() != Mover::FLIP){
		  cerr<<"*error* insertFlip(rblk, "<< m->type() <<")"<<endl;
		  return Seq<char>();
	       }
	       ((FlipMemo*)m)->insertAtom(*atsq);
	    }
	 }
      }
   }
   return resAlts;
}

// Insert an atom into one or more flip memos based on the residues alt codes

void AtomPositions::insertFlip(const PDBrec& hr, Seq<char> alts) {
   char descrbuf[30];

   if (! visableAltConf(hr, _onlyA)) { return; }

   while(alts) {
      if (hr.alt() == ' ' || hr.alt() == *alts) {
	 ::sprintf(descrbuf, "%c%4d%c%-3.3s%-4.4s%c",
	    hr.chain(), hr.resno(), hr.insCode(),
	    hr.resname(), "", (*alts));
	 Stringclass descr = Stringclass(descrbuf);

	 Mover *m = _motionDesc.get(descr);
	 if (m == NULL) {
	    m = new FlipMemo(hr.resname(), _useXplorNames);
	    _motionDesc.put(descr, m);
	 }
	 else if (m->type() != Mover::FLIP){
	    cerr<<"*error* insertFlip(hr, "<< m->type() <<")"<<endl;
	    return;
	 }
	 ((FlipMemo*)m)->insertAtom(hr);
      }
      alts++;
   }
}
// ---------------------------------------------------------------
void AtomPositions::finalizeMovers() {
   DictIterator<Stringclass, Mover> it(_motionDesc);
   Stringclass descr;

   while (it.next(descr)) {
      Mover *m = _motionDesc.get(descr);
      if (m != NULL) {
	 m->descr(descr);
	 m->finalize(_nBondCutoff, _useXplorNames, *this, _dotBucket);
      }
   }
}
// ---------------------------------------------------------------
CliqueList AtomPositions::findCliques() const {
   const float prRad = 0.0; // we just want to examine bumps here
   const int mdsz = _motionDesc.size();
   float maxVDWrad = 1.0;

   MultiDict<LocBlk, BumperPoint> bumpbins(2*mdsz);
   Vector< MoverPtr >   memo(mdsz);
   DictIterator<Stringclass, Mover> it(_motionDesc);
   Stringclass descr;

// pack each memo into an array and add bumper points to bumpbins
   int rn=0, an=0;
   while (it.next(descr)) {
      Mover   *mx = _motionDesc.get(descr);
      if (mx != NULL && mx->valid() && mx->isComplete()
                     && rn < mdsz) {
	 memo[rn] = mx;
#ifdef DEBUGCLIQUE
	 _os << "#res " << rn << " is " << mx->descr() << endl;
#endif
	 an += mx->makebumpers(bumpbins, rn, maxVDWrad);
	 rn++;
      }
   }
   const int numMemos = rn;
   memo.resize(numMemos);

// exclude bumper points which interact with exclude-list members

   for (Seq <BumperPoint> xp = _excludePoints; xp; xp++) {
      Seq<BumperPoint> nearxp = ::neighbors((*xp).loc(), (Coord)0.001,
	       (Coord)((*xp).vdwRad() + prRad + maxVDWrad), bumpbins);
      while (nearxp) {
	 if ((*nearxp).valid()
	    && ((distance((*xp).loc(), (*nearxp).loc())
	           - (*xp).vdwRad() - (*nearxp).vdwRad()) < prRad)) {

	    nearxp.hd().invalidate(); // turn off if bumps exclude pt
	 }
	 nearxp++;
      }
   }

   DisjointSets connsets(numMemos);
   MultiDictIterator<LocBlk, BumperPoint> bbit(bumpbins);

// add connections between memos to DisjointSet (& remember pairwise links)

   LocBlk key;
   while (bbit.next(key)) {
      for (Seq <BumperPoint> bp = bumpbins.get(key); bp; bp++) {
	 Seq<BumperPoint> nearbp = ::neighbors((*bp).loc(), (Coord)0.001,
	       (Coord)((*bp).vdwRad() + prRad + maxVDWrad), bumpbins);
	 while (nearbp) {
	    const int ares = (*bp).rnum(), bres = (*nearbp).rnum();
	    if ((ares != bres) && (*bp).valid() && (*nearbp).valid()
	    && (!sameres(memo[ares]->exampleAtom(), memo[bres]->exampleAtom()))
	    && ((distance((*bp).loc(), (*nearbp).loc())
	           - (*bp).vdwRad() - (*nearbp).vdwRad()) < prRad)) {

#ifdef DEBUGCLIQUE
//	       _os << "-connect " << ares << " to " << bres << endl;
#endif
	       connsets.connect(ares, bres); // build connected sets

	       // remember who connects to whom
	       if (! memo[ares]->linkExists(memo[bres]->descr())) {
		  memo[ares]->addLink(memo[bres]->descr(), memo[bres]);
	       }
	       if (! memo[bres]->linkExists(memo[ares]->descr())) {
		  memo[bres]->addLink(memo[ares]->descr(), memo[ares]);
	       }
	    }
	    nearbp++;
	 }
      }
   }
   
// figure out subset connectivity

   int** djss = connsets.subsets();

// package connectivity info into a CliqueList
   int i = 0, j = 0;

   CliqueList clst;
   for(j=1; j <= djss[0][0]; j++) {
      Seq<MoverPtr> c;
      for(i=1; i <= djss[j][0]; i++) {
         c.insert(memo[ djss[j][i] ]);
      }
      clst.insertClique( mpSeqSort(c) );
   }
   freeDJsubsets(djss); // finished with array of subset indexes

// group non-interacting elements into a single list

   for(i=0; i < numMemos; i++) {
      if (connsets.singleton(i)) { clst.insertSingleton(memo[i]); }
   }
   clst.sortSingletonsByDescr();

#ifdef DEBUGCLIQUE
   _os << "Searched " << rn << " residues and examined " << an << " frontier points." << endl;
#endif

   return clst;
}

// ---------------------------------------------------------------
int AtomPositions::orientSingles(Seq<MoverPtr> singles) {
#ifndef OLD_CXX_DEFNS
   ios_base::fmtflags lff = _os.setf(ios_base::fixed, ios_base::floatfield);
#else
           const long lff = _os.setf(     ios::fixed,      ios::floatfield);
#endif
#ifndef OLD_CXX_DEFNS
   ios_base::fmtflags laf = _os.setf(ios_base::right, ios_base::adjustfield);
#else
           const long laf = _os.setf(     ios::right,      ios::adjustfield);
#endif
   const int iw = _os.width(8);
   const int ip = _os.precision(3);

   int i;

   double penalty = 0.0;
   int numInNonDefaultOrientations = 0;

   float bumpTrialScore = 0.0;
   float hbTrialScore   = 0.0;
   bool trialBadBump    = FALSE;

   while (singles) {
      MoverPtr m = *singles++;
      if (m && m->valid()) {
	 double prevPenalty = 0.0;
	 float theBumpScore = 0.0;
	 float theHBScore   = 0.0;
	 bool hasBadBump    = FALSE;
	 const int numLowResO = m->numOrientations();
	 m->setBestOrientation(0, LowestMoverScore, hasBadBump);

	 // try each orientation
	 for (i = 0; i < numLowResO; i++) {
	    if (m->setOrientation(i, *this)) {
	       const double val = m->determineScore(*this,
	                   _dotBucket, _nBondCutoff, _probeRadius,
			   _pmag, penalty, bumpTrialScore,
			   hbTrialScore, trialBadBump);
	       if (val+penalty > m->bestScore()+prevPenalty) {
		  m->setBestOrientation(i, val, trialBadBump);
		  theBumpScore = bumpTrialScore;
		  theHBScore   = hbTrialScore;
		  hasBadBump   = trialBadBump;
		  prevPenalty  = penalty;
	       }
	       if (_outputNotice && _showOrientScore) {
		  _os << m->descr()
	            << "[" << i+1 << ":" << m->describeOrientation()
		    <<"] bump=" << bumpTrialScore
		    << ", HB=" << hbTrialScore
		    << ", penalty=" << penalty
		    << ", bestScore=" << m->bestScore()
		    << ((trialBadBump==TRUE)?", BADBUMP":"")
		    << endl;
	       }
	       if (m->canFlip()) {
		  m->trackFlipMaxScore(m->flipState(), val, trialBadBump);
	       }
	    }
	    else { break; } // could not set orientation (we are done)
	 }
	 
	 // remember and position to the best
	 const int best = m->bestOrientation();
	 if (m->orientation() != best) {
	    m->setOrientation(best, *this);
	 }
	 if (m->hasHires()) {
	    const int numHiResO = m->numOrientations(Mover::HIGH_RES);
	    // try each high-res orientation
	    for (i = 0; i < numHiResO; i++) {
	       if (m->setOrientation(i, *this, Mover::HIGH_RES)) {
		  const double val = m->determineScore(*this,
	                   _dotBucket, _nBondCutoff, _probeRadius,
			   _pmag, penalty, bumpTrialScore,
			   hbTrialScore, trialBadBump);

		  if ((i == 0) || (val+penalty > m->bestScore()+prevPenalty)) {
		     m->setBestOrientation(i, val, trialBadBump, Mover::HIGH_RES);
		     theBumpScore = bumpTrialScore;
		     theHBScore   = hbTrialScore;
		     hasBadBump   = trialBadBump;
		     prevPenalty  = penalty;
		  }
		  if (_outputNotice && _showOrientScore) {
		     _os << ">" << m->descr()
			 << "[" << best+1 << "." << i+1 << ":"
			 << m->describeOrientation()
			 <<"] bump=" << bumpTrialScore
			 << ", HB=" << hbTrialScore
			 << ", penalty=" << penalty
			 << ", bestScore=" << m->bestScore()
			 << ((trialBadBump==TRUE)?", BADBUMP":"")
			 << endl;
		  }
		  if (m->canFlip()) {
		     m->trackFlipMaxScore(m->flipState(), val, trialBadBump);
		  }
	       }
	       else { break; } // could not set orientation (we are done)
	    }
	 
	    // remember and position to the best
	    const int bestHires = m->bestOrientation(Mover::HIGH_RES);
	    if (m->orientation(Mover::HIGH_RES) != bestHires) {
	       m->setOrientation(bestHires, *this, Mover::HIGH_RES);
	    }
	    if (! m->isDefaultO(bestHires, Mover::HIGH_RES)) {
	       numInNonDefaultOrientations++;
	    }
	 }
	 else {
	    if (! m->isDefaultO(best)) {
	       numInNonDefaultOrientations++;
	    }
	 }
	 if (m->canFlip() && (m->flipState() != 0)) {
	    m->markFlipAtoms();
	 }
	 if (_outputNotice) {
	    _os << " orientation " << best+1
	        << ":" << m->descr()
	        << ":" << m->describeOrientation()
	        << ": bump=" << theBumpScore
		<< ", HB=" << theHBScore
		<< ", total=" << m->bestScore()
	        << ((hasBadBump==TRUE)?", BADBUMP":"")
		<< endl;
	 }
      }
   }
#ifndef OLD_CXX_DEFNS
   _os.setf(lff, ios_base::floatfield);
   _os.setf(laf, ios_base::adjustfield);
#else
   _os.setf(lff, ios::floatfield);
   _os.setf(laf, ios::adjustfield);
#endif
   _os.width(iw);
   _os.precision(ip);
   return numInNonDefaultOrientations;
}

// ---------------------------------------------------------------
int AtomPositions::orientClique(Seq<MoverPtr> clique, int limit) {
#ifndef OLD_CXX_DEFNS
   ios_base::fmtflags lff = _os.setf(ios_base::fixed, ios_base::floatfield);
#else
           const long lff = _os.setf(     ios::fixed,      ios::floatfield);
#endif
#ifndef OLD_CXX_DEFNS
   ios_base::fmtflags laf = _os.setf(ios_base::right, ios_base::adjustfield);
#else
           const long laf = _os.setf(     ios::right,      ios::adjustfield);
#endif
   const int iw = _os.width(8);
   const int ip = _os.precision(3);

   double log10ncombo = 0.0;
   int i = 0, rc = 0;

   if (_outputNotice) { _os << " Processing set"; }
   for(Seq<MoverPtr> s = clique; s; s++) {
      const int no = (*s)->numOrientations();
      log10ncombo += log10(double(no));
      if (_outputNotice && ++i > 3) { i = 0; _os << endl<< "  "; }
      if (_outputNotice) { _os << ":" << (*s)->descr() << "[" << no << "]"; }
   }
   const double p10 = floor(log10ncombo);
   const float mantis = pow(double(10.0), log10ncombo - p10);
   const int poweroften = int(p10);
   if(_outputNotice){
      if (log10ncombo < log10(double(1000000))) { //smallish numbers as integers
      const long ncombo = (long)floor(pow(double(10.0), log10ncombo)+0.5);
      _os << ": "<< ncombo << " permutations." << endl;
      }
      else {
      _os << ": "<< mantis <<"E" << poweroften << " permutations." << endl;
      }
   }
#ifdef DEBUG
   cerr<<"DEBUG: clique.len() = "<<clique.len()<<endl;
#endif
   if (log10ncombo < log10(double(limit))) {
      rc = exhaustiveSearchOfClique(clique);
   }
   else {
      cerr << endl << "*** PROCESSING OF SET ABANDONED! - Can't search "
	 << mantis <<"E" << poweroften << " states! ***" << endl << endl;

//			PDBrec thisAtom;
//			Seq<MoverPtr> s = clique;
//			thisAtom = (*s)->exampleAtom();
//		  cerr <<"DEBUG Bad Atom:  " << PDBrecNAMEout(thisAtom)<<endl;

		 for(Seq<MoverPtr> s = clique; s; s++) {
		    cerr << (*s)->descr() << endl;
		 }	
				
      cout << "ERROR Set not processed by reduce. Requires "
	 << mantis <<"E" << poweroften << " states!" << endl;
      rc = -1; // problem signal
   }
#ifndef OLD_CXX_DEFNS
   _os.setf(lff, ios_base::floatfield);
   _os.setf(laf, ios_base::adjustfield);
#else
   _os.setf(lff, ios::floatfield);
   _os.setf(laf, ios::adjustfield);
#endif
   _os.width(iw);
   _os.precision(ip);

   return rc;
}
// ---------------------------------------------------------------
int AtomPositions::exhaustiveSearchOfClique(Seq<MoverPtr> clique) {
   const int numItems = clique.len();
   Vector<MoverPtr>    item(numItems);
   Vector<double> currScore(numItems);
   Vector<float>   currBump(numItems);
   Vector<float>  currHbond(numItems);
   Vector<bool> currBadBump(numItems);
   double scoreThisO       = 0.0;
   double penalty          = 0.0;
   double worstPenalty     = 0.0;
   double prevWorstPenalty = 0.0;
   int i = 0, best = 0;

   // ---------------------------------------------------------------
   // initialize
   for (i = 0; clique; clique++, i++) {
      item[i] = *clique;
      if (! item[i]->initOrientation(*this)) {
	 cerr << endl << "ERROR: exhaustiveSearchOfClique: failed to initialize item - "
	              << item[i]->descr() << endl;
	 return 0;
      }
      item[i]->setBestOrientation(0, LowestMoverScore, FALSE);
   }

   // ---------------------------------------------------------------
   // search through each permutation

   double bestOScore = LowestMoverScore;

   int cursor = 0;
   bool overflow = FALSE;
   while (cursor < numItems) {
      while(! overflow) {
	 scoreThisO   = 0.0;
	 worstPenalty = 0.0;
	 for (i = 0; i < numItems; i++) {
	    currScore[i] = item[i]->determineScore(*this,
	                   _dotBucket, _nBondCutoff, _probeRadius,
			   _pmag, penalty, currBump[i],
			   currHbond[i], currBadBump[i]);
	    if (penalty < worstPenalty) { worstPenalty = penalty; }

	    if (_outputNotice) {
	       if (_showOrientScore) {
		  _os <<item[i]->descr()
		      << "[" << item[i]->orientation()+1
		      << ":" << item[i]->describeOrientation()
		      << "] bump=" << currBump[i]
		      << ", HB=" << currHbond[i]
		      << ((currBadBump[i]==TRUE)?", BADBUMP":"")
		      << endl;
	       }
	       else if (_cliqueTicks) {
		  _os <<item[i]->orientation()+1<<" ";
	       }
	    }
	    scoreThisO += currScore[i];
	 }
	 for (i = 0; i < numItems; i++) {
	    if (item[i]->canFlip()) {
	       item[i]->trackFlipMaxScore(item[i]->flipState(),
	                                  scoreThisO, currBadBump[i]);
	    }
	 }

	 if (scoreThisO+worstPenalty > bestOScore+prevWorstPenalty) {
	    bestOScore = scoreThisO;
	    prevWorstPenalty = worstPenalty;
	    for (i = 0; i < numItems; i++) {
	       item[i]->setBestOrientation(
	              item[i]->orientation(),
		      currScore[i], currBadBump[i]);
	    }
	 }
	 if (_outputNotice) {
	    if (_showOrientScore) {
	       _os << " permutation score =" << scoreThisO
	           << ", maxPenalty=" << worstPenalty
	           << ", bestScore=" << bestOScore << endl;
	    }
	    else if (_cliqueTicks) {
	       _os << " : " << bestOScore << "       \r";
	    }
	 }
	 if (! item[cursor=0]->nextOrientation(*this)) {
	    overflow = TRUE;
	 }
      }

      item[cursor]->initOrientation(*this);
      overflow = FALSE;

      if (++cursor < numItems) {
	 if (! item[cursor]->nextOrientation(*this)) {
	    overflow = TRUE;
	 }
      }
   }
   // ---------------------------------------------------------------
   // re-position to the orientations with the best clique scores
   for (i = 0; i < numItems; i++) {
      best = item[i]->bestOrientation();
      if (item[i]->orientation() != best) {
	 item[i]->setOrientation(best, *this);
      }
   }
   // ---------------------------------------------------------------
   // optimization of any hires movers and get total score

   int numInNonDefaultOrientations = 0;

   for (cursor = 0; cursor < numItems; cursor++) {
      if (item[cursor]->hasHires()) {
	 const int numHiResO = item[cursor]->numOrientations(Mover::HIGH_RES);
	 // try each high-res orientation
	 for (int ho = 0; ho < numHiResO; ho++) {
	    if (item[cursor]->setOrientation(ho, *this, Mover::HIGH_RES)) {
	       scoreThisO   = 0.0;
	       worstPenalty = 0.0;
	       if (_outputNotice && (! _showOrientScore) && _cliqueTicks){
		  _os << ">";
	       }
	       for (i = 0; i < numItems; i++) {
 		  currScore[i] = item[i]->determineScore(*this,
	                   _dotBucket, _nBondCutoff, _probeRadius,
			   _pmag, penalty, currBump[i],
			   currHbond[i], currBadBump[i]);
		  if (penalty < worstPenalty) { worstPenalty = penalty; }
		  if (_outputNotice) {
		     if (_showOrientScore) {
			_os << ">" <<item[i]->descr()
			   << "[" << item[i]->orientation()+1;
			if (i == cursor) { _os << "." << ho+1; }
			_os << ":" << item[i]->describeOrientation()
			   << "] bump=" << currBump[i]
			   << ", HB=" << currHbond[i]
			   << ((currBadBump[i]==TRUE)?", BADBUMP":"")
			   << endl;
		     }
		     else if (_cliqueTicks) {
			_os <<item[i]->orientation()+1;
			if (i == cursor) { _os <<"."<<ho+1; }
			_os <<" ";
		     }
		  }
		  scoreThisO += currScore[i];
	       }
	       for (i = 0; i < numItems; i++) {
		  if (item[i]->canFlip()) {
		     item[i]->trackFlipMaxScore(item[i]->flipState(),
	                                  scoreThisO, currBadBump[i]);
		  }
	       }

	       if ((ho == 0) || (scoreThisO+worstPenalty > bestOScore+prevWorstPenalty)) {
		  bestOScore = scoreThisO;
		  prevWorstPenalty = worstPenalty;
		  for (i = 0; i < numItems; i++) {
		     item[i]->setBestOrientation(
			item[i]->orientation(Mover::HIGH_RES),
			currScore[i], currBadBump[i], Mover::HIGH_RES);
		  }
	       }
	       if (_outputNotice) {
		  if (_showOrientScore) {
		    _os << "> adjustment score =" << scoreThisO
			<< ", maxPenalty=" << worstPenalty
			<< ", bestScore=" << bestOScore << endl;
		  }
		  else if (_cliqueTicks) {
		     _os << " : " << bestOScore << "       \r";
		  }
	       }
	    }
	    else { break; } // could not set orientation (we are done)
	 }
	 // remember and position to the best
	 const int bestHires = item[cursor]->bestOrientation(Mover::HIGH_RES);
	 if (item[cursor]->orientation(Mover::HIGH_RES) != bestHires) {
	    item[cursor]->setOrientation(bestHires, *this, Mover::HIGH_RES);
	 }
	 if (! item[cursor]->isDefaultO(bestHires, Mover::HIGH_RES)) {
	    numInNonDefaultOrientations++;
	 }
	 currScore[cursor] = item[cursor]->determineScore(*this,
	                   _dotBucket, _nBondCutoff, _probeRadius,
			   _pmag, penalty, currBump[cursor],
			   currHbond[cursor], currBadBump[cursor]);
      }
      else {
	 best = item[cursor]->bestOrientation();
	 if (! item[cursor]->isDefaultO(best)) {
	    numInNonDefaultOrientations++;
	 }
      }
   }
   // ---------------------------------------------------------------
   // tally the results

   double totScore = 0.0;

   for (i = 0; i < numItems; i++) {
      const int best = item[i]->bestOrientation();
      if (item[i]->canFlip() && (item[i]->flipState() != 0)) {
	 item[i]->markFlipAtoms();
      }
      if (_outputNotice) {
	 _os << " orientation " << best+1
	        << ": " << item[i]->descr()
	        << ": " << item[i]->describeOrientation()
	        << ":bump=" << currBump[i]
		<< ", HB=" << currHbond[i]
	        << ((currBadBump[i]==TRUE)?", BADBUMP":"")
		<< endl;
      }
      totScore += item[i]->bestScore();
   }

   if (_outputNotice) {
      _os << " Total score for set: " << totScore << endl;
   }

   return numInNonDefaultOrientations;
}

// ---------------------------------------------------------------
// insert user records about motion at the end of the header
void AtomPositions::describeChanges(NonConstListIter<PDBrec>& infoPtr,
                                    List<Stringclass>& notes) {
   if (notes.size() > 0) {
      PDBrec separator("USER  MOD -----------------------------------------------------------------");
      infoPtr.insertBefore(separator);
      infoPtr.insertBefore(PDBrec(
	 "USER  MOD scores for adjustable sidechains, with \"set\" totals for H,N and Q"));
      infoPtr.insertBefore(PDBrec(
	 "USER  MOD \"o\" means original, \"f\" means flipped, \"180deg\" is methyl default"));

      char fmtbuf[123];
      ::sprintf(fmtbuf, "USER  MOD \"!\" flags a clash with an overlap of %.2fA or greater",
						_bad_bump_gap_cutoff);
      infoPtr.insertBefore(PDBrec(fmtbuf));

      ::sprintf(fmtbuf, "USER  MOD flip categories: \"K\"=keep, \"C\"=clashes, \"X\"=uncertain, \"F\"=flip");
      infoPtr.insertBefore(PDBrec(fmtbuf));

      Stringclass s;

      NonConstListIter<Stringclass> np(notes);

      while(np.next(s)) {	// insert notes in sorted order
	 PDBrec userRec(s.array());
	 infoPtr.insertBefore(userRec);
      }

      infoPtr.insertBefore(separator);
   }
}

// ---------------------------------------------------------------
void AtomPositions::manageMetals(const ResBlk& rblk) {
   MultiDictIterator<Stringclass, PDBrec> pdbit = rblk.atomIt();
   Stringclass key, descr;

   while(pdbit.next(key)) {
      for (Seq<PDBrec> a = rblk.get(key); a; a++) {
	 if ((*a).hasProp(METALIC_ATOM)) {
	    _excludePoints.insert(BumperPoint((*a).loc(), 0, 0, (*a).vdwRad()));
	 }
      }
   }
}

// ---------------------------------------------------------------
// create possible orientations for H atoms on waters (identified elsewhere)
// and store these Hs in the xyz table
void AtomPositions::generateWaterPhantomHs(Seq<PDBrec>& waters) {
   char descrbuf[30];
   const int MaxAccDir = 25;
   struct AccDirection {
      AccDirection() : _nam(""), _gap(999.9) {}
      AccDirection(const Stringclass& s, const Point3d& p, float g)
	 : _nam(s), _loc(p), _gap(g) {}
      Stringclass  _nam;
      Point3d _loc;
      float   _gap;
   } nearbyA[MaxAccDir];

  const ElementInfo& elemHOd = * ElementInfo::StdElemTbl().element("HOd");

  for(Seq<PDBrec>a = waters; a; a++) {
    int i = 0, nAcc = 0;

    for(Seq<PDBrec> nearby = neighbors((*a).loc(), 0.001, 4.0);
	     nearby; nearby++) {

      if ((*nearby).hasProp(ACCEPTOR_ATOM)
      || FlipMemo::isHBDonorOrAcceptorFlipped(*nearby, _useXplorNames)) {

        if (abs((*nearby).occupancy()) > _occupancyCuttoff
          && interactingConfs(*a, *nearby, _onlyA)) {
          double HBoverlap = distance((*a).loc(), (*nearby).loc())
                             - ( elemHOd.explRad() + (*nearby).vdwRad()
                             + 1.0 /*i.e. the O-H bond length*/ );

          if (HBoverlap < -0.01) {

	 // Now we make a table of all the locations of each HB
	 // acceptor nearby keeping ONLY one for each aromatic ring
	 // (the one with the most negative gap).

            bool foundMatchingAromAtom = FALSE;
            bool isAromRingAtom = StdResH::ResXtraInfo().atomHasAttrib(
                         (*nearby).resname(), (*nearby).atomname(), AROMATICFLAG);

            ::sprintf(descrbuf, "%-3.3s%c%-3.3s%4d%c%c",
                                (isAromRingAtom ? "/R/" : ""), (*nearby).alt(),
                                (*nearby).resname(), (*nearby).resno(),
                                (*nearby).insCode(), (*nearby).chain());
            Stringclass resDescr = Stringclass(descrbuf);

            if (isAromRingAtom) {
              for (i=0; i < nAcc; i++) {
                if (resDescr == nearbyA[i]._nam) {
                  if (HBoverlap < nearbyA[i]._gap) {
                    nearbyA[i]._nam = resDescr;
                    nearbyA[i]._loc = (*nearby).loc();
                    nearbyA[i]._gap = HBoverlap;
					   }
                  foundMatchingAromAtom = TRUE;
                  break; // for i<nAcc
				    }
			     } // for i<nAcc
			   }
	         if (!foundMatchingAromAtom && nAcc < MaxAccDir) {
		        nearbyA[nAcc]._nam = resDescr;
		        nearbyA[nAcc]._loc = (*nearby).loc();
		        nearbyA[nAcc]._gap = HBoverlap;
		        nAcc++;
	         }
          } // if HBoverlap
	     } // if occupancy good
      } // if acceptor
    } // for nearby
  // each of these acceptors is a possible direction for a hydrogen

    for(i = 0; i < nAcc; i++) {
	   PDBrec pHatom = (*a).clone(); // duplicate & modify Oxygen

#define BEST_HBOND_OVERLAP 0.6
	   double waterOHbondLen = 1.0 +
								 max(-1.0, min(0.0, nearbyA[i]._gap + BEST_HBOND_OVERLAP));

	   pHatom.loc((nearbyA[i]._loc - (*a).loc()).scaleTo(waterOHbondLen)
	                                             + (*a).loc());
	   pHatom.atomname(" H? ");

	   pHatom.elem(elemHOd);

	   pHatom.atomno(0);
	   pHatom.occupancy(_occupancyCuttoff + 0.01);
	   pHatom.tempFactor(99.99);
	   pHatom.elemLabel(" H");
	   pHatom.chargeLabel("  ");
	   pHatom.annotation("dummy atom");

	   put(pHatom); // index in the xyz table
#ifdef DUMPPHANH
//if (waterOHbondLen < 1.0) 
      _os << "{" << PDBrecNAMEout(*a) << "}P "
	       << (*a).x()<< ", " << (*a).y() << ", " << (*a).z()
	       << endl
	       << "{" << PDBrecNAMEout(pHatom) << "}L "
	       << pHatom.x()<< ", " << pHatom.y() << ", " << pHatom.z()
	       << endl;
#endif
    } // for nAcc
  } // for waters
}

// ---------------------------------------------------------------
// score for a single atom a at point p based on interactions with
// nearby atoms (ignoring excluded atoms)

double AtomPositions::atomScore(const PDBrec& a, const Point3d& p,
 				const Seq<PDBrec>& nearby, const Seq<PDBrec>& exclude,
				const DotSph& dots, float pRad,
				bool onlyBumps,
				float &bumpSubScore, float &hbSubScore,
				bool &hasBadBump) {
  bumpSubScore = 0.0;
  hbSubScore   = 0.0;
  hasBadBump   = FALSE;

  if ((! a.valid()) || a.hasProp(IGNORE)
  ||  (! visableAltConf(a, _onlyA))
  ||  (abs(a.occupancy()) <= _occupancyCuttoff) ) {
    return 0.0;
  }

  Seq<PDBrec> bumping; // first we collect atoms actually interacting

  for (Seq_item<PDBrec>* ni = nearby.start(); ni; ni = Seq<PDBrec>::next(ni)) {
	  if ((ni->data()).valid() && (! (ni->data()).hasProp(IGNORE))
	  && (abs((ni->data()).occupancy()) > _occupancyCuttoff)
	  && (! (a == ni->data()))
	  && (interactingConfs(a, ni->data(), _onlyA))
	  && ( vdwGap(a, p, ni->data(), (ni->data()).loc()) < pRad) 
	  && (! foundInList(ni->data(), exclude))) {
	    bumping.insert(ni->data());
	  }
  }

  int HBmask = 0;
  if (a.hasProp(   DONOR_ATOM)) { HBmask |= ACCEPTOR_ATOM; }
  if (a.hasProp(ACCEPTOR_ATOM)) { HBmask |= DONOR_ATOM;    }

  const int ndots = dots.count();
  double s = 0.0;
  for (int i=0; i < ndots; i++) {  // then we inspect each dots interactions
    const Point3d q = p + dots[i];

    const Point3d probeLoc = (pRad > 0.0)
                             ? p + dots[i].scaled(dots.radius()+pRad)
                             : q;

    double mingap = 999.9;
	 bool keepDot = FALSE;
	 bool isaHB = FALSE;
	 bool tooCloseHB = FALSE;
	 Seq_item<PDBrec>* cause;
	 float HBmindist = 999.9;

    for (Seq_item<PDBrec>* b = bumping.start(); b; b = Seq<PDBrec>::next(b)) {
	   const Point3d locb = (b->data()).loc();
	   const double vdwb = (b->data()).vdwRad();

	   const double probeGap = distance( probeLoc, locb ) - pRad - vdwb;
	   const double      gap = distance( q, locb )               - vdwb;

	   if (probeGap < 0.0) {
	     if (gap < mingap) {
		    const bool bothCharged = a.isCharged() && (b->data()).isCharged();

		    const bool chargeComplement = bothCharged
                  ?  ( (a.isPositive() && (b->data()).isNegative()) ||
                       (a.isNegative() && (b->data()).isPositive()) )
		            : FALSE;

          if( (b->data()).hasProp(HBmask) &&
                   ((! bothCharged) || chargeComplement)) { // H-bond
		      isaHB = TRUE;

            HBmindist = bothCharged ?
                        _min_charged_hb_cutoff :
                        _min_regular_hb_cutoff;

		      tooCloseHB = (gap < -HBmindist);
		    }
		    else { // clash or contact
		      if( (b->data()).hasProp(HB_ONLY_DUMMY) ) {
		        continue; // *** dummy only counts as an HBond partner
		      }
		      isaHB = tooCloseHB = FALSE;
		    }
		    cause = b;
		    keepDot = TRUE;
		    mingap = gap;
		  }
	   }
    }
    if (keepDot) { // remove dots inside a connected atom
      for (Seq_item<PDBrec>* x = exclude.start(); x; x = Seq<PDBrec>::next(x)) {
	     if ( (x->data()).valid() && (! (x->data()).hasProp(IGNORE))
		  && interactingConfs(a, x->data(), _onlyA) 
		  && (distance(q, (x->data()).loc()) < (x->data()).vdwRad())) {
	       keepDot = FALSE;
	       break;
	     }
	   }
    }

    const float GAPscale  = 0.25; // GAP score scale Factor
    const float BUMPweight =10.0; // BUMP score weight
    const float HBweight   = 4.0; // HBond score weight

    if (keepDot) { // determine partial dot scores

	   int overlapType = 0;
	   double overlap = 0.0;

      if (mingap > 0.0)      {overlap = 0.0;       overlapType =  0;}
      else if (isaHB && tooCloseHB) {
        mingap += HBmindist;
		  overlap =-0.5*mingap;
		  overlapType = -1;
	   }
      else if (isaHB)        {overlap =-0.5*mingap;overlapType =  1;}
      else if (mingap < 0.0) {overlap =-0.5*mingap;overlapType = -1;}

#ifdef DEBUGCONTACT
      if(strstr(":ASN:", a.resname()) && a.resno() == 364)
        _os << "*contact :" << PDBrecNAMEout(a) << ":" <<  PDBrecNAMEout(cause->data())
            << ": type=" << overlapType << ", mingap=" <<  mingap
            << ", q@" << q << ", cause@" << cause->data().loc() << endl;
#endif
	 // using scale values from probe program
      if (overlapType == -1) {              // clash
	     const double bmp = -BUMPweight*overlap;
	     bumpSubScore += bmp;
	     s += bmp;
	     if (mingap <= -_bad_bump_gap_cutoff) { hasBadBump = TRUE; }
      }
      else if (overlapType == 0) {			    // contact dot
	     if ((!onlyBumps)
	     && ! annularDots(q, a, cause->data(), pRad)) {
	       const double scaledGap = mingap/GAPscale;
	       s += exp(-scaledGap*scaledGap);
	    }
      }
      else if (overlapType ==  1) {			    // H-bond
        if (!onlyBumps) {
	       const double hbv = HBweight*overlap;
	       hbSubScore += hbv;
	       s += hbv;
        }
        else {   // in this case treat as a bump
	       const double pseudobump = -BUMPweight*overlap;
	       bumpSubScore += pseudobump;
	       s += pseudobump;
        }
      } // else if
    }
  } // if keepDot
  bumpSubScore /= dots.density();
  hbSubScore /= dots.density();
  s /= dots.density();
#ifdef DEBUGATOMSCORE
  _os << "\t:" << PDBrecNAMEout(a)
	   << ": bump=" << bumpSubScore
	   << ", HB=" << hbSubScore
	   << ((hasBadBump==TRUE)?", BADBUMP":"")
		<<", partialScore=" << s
	   << endl;
#endif
  return s;
}
